from django.apps import AppConfig
from django.core import checks

from .checks import storage_check


class S3FileConfig(AppConfig):
    name = 's3file'
    verbose_name = 'S3File'

    def ready(self):
        from django.core.files.storage import default_storage
        from storages.backends.s3boto3 import S3Boto3Storage
        from django import forms

        from .forms import S3FileInputMixin

        if isinstance(default_storage, S3Boto3Storage) and \
                S3FileInputMixin not in forms.ClearableFileInput.__bases__:
            forms.ClearableFileInput.__bases__ = \
                (S3FileInputMixin,) + forms.ClearableFileInput.__bases__

        elif S3FileInputMixin in forms.ClearableFileInput.__bases__:
            forms.ClearableFileInput.__bases__ = tuple(
                cls for cls in forms.ClearableFileInput.__bases__
                if cls is not S3FileInputMixin
            )

        checks.register(storage_check, checks.Tags.security, deploy=True)



import base64
import hashlib
import hmac
import logging

from django import http
from django.conf import settings
from django.core.files.storage import default_storage
from django.views import generic

logger = logging.getLogger('s3file')


class S3MockView(generic.View):

    def post(self, request):
        success_action_status = request.POST.get('success_action_status', 201)
        try:
            file = request.FILES['file']
            key = request.POST['key']
            date = request.POST['x-amz-date']
            signature = request.POST['x-amz-signature']
            policy = request.POST['policy']
        except KeyError:
            logger.exception('bad request')
            return http.HttpResponseBadRequest()

        try:
            signature = base64.b64decode(signature.encode())
            policy = base64.b64decode(policy.encode())

            calc_sign = hmac.new(
                settings.SECRET_KEY.encode(),
                policy + date.encode(),
                'sha256'
            ).digest()
        except ValueError:
            logger.exception('bad request')
            return http.HttpResponseBadRequest()

        if not hmac.compare_digest(signature, calc_sign):
            logger.warning('bad signature')
            return http.HttpResponseForbidden()

        key = key.replace('${filename}', file.name)
        etag = hashlib.md5(file.read()).hexdigest()  # nosec
        file.seek(0)
        key = default_storage.save(key, file)
        return http.HttpResponse(
            "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
            "<PostResponse>"
            f"<Location>{settings.MEDIA_URL}{key}</Location>"
            f"<Bucket>{getattr(settings, 'AWS_STORAGE_BUCKET_NAME')}</Bucket>"
            f"<Key>{key}</Key>"
            f"<ETag>\"{etag}\"</ETag>"
            "</PostResponse>",
            status=success_action_status,
        )



from django.core.checks import Error
from django.core.files.storage import FileSystemStorage, default_storage


def storage_check(app_configs, **kwargs):
    if isinstance(default_storage, FileSystemStorage):
        return [
            Error(
                'FileSystemStorage should not be used in a production environment.',
                hint='Please verify your DEFAULT_FILE_STORAGE setting.',
                id='s3file.E001',
            )
        ]
    return []



import base64
import logging
import pathlib
import uuid

from django.conf import settings
from django.utils.functional import cached_property
from storages.utils import safe_join

from s3file.storages import storage

logger = logging.getLogger('s3file')


class S3FileInputMixin:
    """FileInput that uses JavaScript to directly upload to Amazon S3."""

    needs_multipart_form = False
    upload_path = getattr(
        settings, 'S3FILE_UPLOAD_PATH', pathlib.PurePosixPath('tmp', 's3file')
    )
    upload_path = safe_join(storage.location, upload_path)
    expires = settings.SESSION_COOKIE_AGE

    @property
    def bucket_name(self):
        return storage.bucket.name

    @property
    def client(self):
        return storage.connection.meta.client

    def build_attrs(self, *args, **kwargs):
        attrs = super().build_attrs(*args, **kwargs)

        accept = attrs.get('accept')
        response = self.client.generate_presigned_post(
            self.bucket_name,
            str(pathlib.PurePosixPath(self.upload_folder, '${filename}')),
            Conditions=self.get_conditions(accept),
            ExpiresIn=self.expires,
        )

        defaults = {
            'data-fields-%s' % key: value
            for key, value in response['fields'].items()
        }
        defaults['data-url'] = response['url']
        defaults.update(attrs)

        try:
            defaults['class'] += ' s3file'
        except KeyError:
            defaults['class'] = 's3file'
        return defaults

    def get_conditions(self, accept):
        conditions = [
            {"bucket": self.bucket_name},
            ["starts-with", "$key", str(self.upload_folder)],
            {"success_action_status": "201"},
        ]
        if accept and ',' not in accept:
            top_type, sub_type = accept.split('/', 1)
            if sub_type == '*':
                conditions.append(["starts-with", "$Content-Type", "%s/" % top_type])
            else:
                conditions.append({"Content-Type": accept})
        else:
            conditions.append(["starts-with", "$Content-Type", ""])

        return conditions

    @cached_property
    def upload_folder(self):
        return str(pathlib.PurePosixPath(
            self.upload_path,
            base64.urlsafe_b64encode(
                uuid.uuid4().bytes
            ).decode('utf-8').rstrip('=\n'),
        ))  # S3 uses POSIX paths

    class Media:
        js = (
            's3file/js/s3file.js' if settings.DEBUG else 's3file/js/s3file.min.js',
        )



import base64
import datetime
import hmac
import json
import os

from django.conf import settings
from django.core.files.storage import FileSystemStorage, default_storage
from django.utils._os import safe_join


class S3MockStorage(FileSystemStorage):
    @property
    def location(self):
        return getattr(settings, 'AWS_LOCATION', '')

    def path(self, name):
        return safe_join(os.path.abspath(self.base_location), self.location, name)

    class connection:
        class meta:
            class client:
                @staticmethod
                def generate_presigned_post(bucket_name, key, **policy):
                    policy = json.dumps(policy).encode()
                    policy_b64 = base64.b64encode(policy).decode()
                    date = datetime.datetime.utcnow().strftime('%Y%m%dT%H%M%SZ')
                    aws_id = getattr(
                        settings, 'AWS_ACCESS_KEY_ID',
                        'AWS_ACCESS_KEY_ID',
                    )
                    fields = {
                        'x-amz-algorithm': 'AWS4-HMAC-SHA256',
                        'x-amz-date': date,
                        'x-amz-credential': aws_id,
                        'policy': policy_b64,
                        'key': key,
                    }
                    signature = hmac.new(
                        settings.SECRET_KEY.encode(),
                        policy + date.encode(),
                        'sha256',
                    ).digest()
                    signature = base64.b64encode(signature).decode()
                    return {
                        'url': '/__s3_mock__/',
                        'fields': {
                            'x-amz-signature': signature,
                            **fields
                        },
                    }

    class bucket:
        name = 'test-bucket'


local_dev = isinstance(default_storage, FileSystemStorage)

storage = default_storage if not local_dev else S3MockStorage()



import json
import logging
import os

from s3file.storages import local_dev, storage

from . import views

logger = logging.getLogger('s3file')


class S3FileMiddleware:

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        file_fields = json.loads(request.POST.get('s3file', '[]'))
        for field_name in file_fields:
            paths = json.loads(request.POST.get(field_name, '[]'))
            request.FILES.setlist(field_name, list(self.get_files_from_storage(paths)))

        if local_dev and request.path == '/__s3_mock__/':
            return views.S3MockView.as_view()(request)

        return self.get_response(request)

    @staticmethod
    def get_files_from_storage(paths):
        """Return S3 file where the name does not include the path."""
        for path in paths:
            path = path.replace(os.path.dirname(storage.location) + '/', '', 1)
            try:
                f = storage.open(path)
                f.name = os.path.basename(path)
                yield f
            except OSError:
                logger.exception("File not found: %s", path)
